function B=chol(A,type)
% chol   Cholesky factorizations of an amat object
% -- B=chol(A) 
% -- B=chol(A,"lower")
% -- B=chol(A,"upper")
%   Compute the upper Cholesky factors, R, of the real symmetric or
%   complex Hermitian positive definite amat A.
%   The upper Cholesky factors R is computed by using the upper
%   triangular part of amat A and is defined by
%  
%          B(k)' * B(k) = A(k), for all k in {1,...,A.N}
%            where A(k) is the k-th matrix in A
%  
%   Calling 'chol' using the optional "upper" flag has the same
%   behavior.  In contrast, using the optional "lower" flag, 'chol'
%   returns the lower triangular factorizations, computed by using the
%   lower triangular part of amat A, such that
%  
%          B(k) * B(k)' = A(k), for all k in {1,...,A.N}.
%
  assert( A.nr == A.nc )
  if nargin==1, type='upper';end
  if strcmp(lower(type),'upper')
    B=amat(cholesky_upper(A.values));return
  end
  if strcmp(lower(type),'lower')
    B=amat(cholesky_lower(A.values));return
  end
  error('second parameter must be ''lower'' or ''upper''!')
end
 
function B=cholesky_lower(A)
% A : N-by-d-by-d, symmetric definite positve
  N=size(A,1);n=size(A,2);
  assert(size(A,3)==n);
  B=zeros(N,n,n);
  
  for j=1:n % Compute the j-th column
    S=0;
    for k=1:j-1
      S=S+abs(B(:,j,k)).^2;
    end
    % test A(:,j,j)-S >= 0 TODO
    B(:,j,j)=sqrt(A(:,j,j)-S);
    for i=j+1:n
      S=0;
      for k=1:j-1
        S=S+B(:,i,k).*conj(B(:,j,k));
      end
      B(:,i,j)=(A(:,i,j)-S)./B(:,j,j);
    end
  end
end

function B=cholesky_upper(A)
% A : N-by-d-by-d, symmetric/hermitian definite positve
  N=size(A,1);n=size(A,2);
  assert(size(A,3)==n);
  B=zeros(N,n,n);
  
  for j=1:n  % Compute the j-th row
    S=0;
    for k=1:j-1
      S=S+abs(B(:,k,j)).^2;
    end
    % test A(:,j,j)-S >= 0 TODO
    B(:,j,j)=sqrt(A(:,j,j)-S);
    for i=j+1:n
      S=0;
      for k=1:j-1
        S=S+B(:,k,i).*conj(B(:,k,j));
      end
      B(:,j,i)=(A(:,j,i)-S)./B(:,j,j);
    end
  end
end
